import ast
import dataclasses
import functools
import json
import logging
import queue
import time
from concurrent.futures import ThreadPoolExecutor
from typing import Any, NamedTuple, Optional

import requests
from opentelemetry import baggage, context, trace

from jf_ingest import logging_helper
from jf_ingest.config import IngestionConfig
from jf_ingest.jf_jira import (
    get_jira_connection,
    get_jira_search_batch_size,
    pull_jira_issues_by_jira_ids,
)
from jf_ingest.telemetry.tracing import jelly_trace, record_span
from jf_ingest.utils import ThreadPoolExecutorWithLogging, retry_for_status

JELLYFISH_CUSTOM_FIELDS_API_LIMIT = 1_000
JELLYFISH_API_TIMEOUT = 600.0


@dataclasses.dataclass
class JCFVDBData:
    field_id: int
    field_key: str
    field_type: str
    value: str


@dataclasses.dataclass
class IssueJCFVDBData:
    issue_jira_id: int
    custom_field_values: list[JCFVDBData]


@dataclasses.dataclass
class JCFVUpdate:
    jira_issue_id: int
    field_id: int
    field_key: str
    field_type: str
    value_old: Optional[str]
    value_new: Optional[str]


@functools.lru_cache(maxsize=4096)
def _literal_eval_memoized(value: str) -> dict:
    """ast.literal_eval is rather slow, and we expect a lot of duplicate values; memoize it."""
    return ast.literal_eval(value)


@functools.lru_cache(maxsize=4096)
def _json_loads_memoized(value: str) -> dict:
    """improve performance of JSON loads by memoizing the results."""
    return json.loads(value)


@jelly_trace
def _annotate_results_from_jellyfish(data: dict) -> dict[str, IssueJCFVDBData]:
    """
    Unpacks the results from Jellyfish into a dict[str, IssueJCFVDBData] - where `str` is the jira_issue_id.

    Typically, the value being sent over is JSON, so we need to handle that here.
    """
    resp = {}
    for issue in data['issues']:
        resp[str(issue['issue_jira_id'])] = IssueJCFVDBData(
            issue_jira_id=issue['issue_jira_id'], custom_field_values=[]
        )
        for jcfv_db_data in issue['custom_field_values']:
            value = jcfv_db_data['value']
            if value and isinstance(value, str):
                try:
                    value = _json_loads_memoized(value)
                except (json.JSONDecodeError, ValueError):
                    try:
                        value = _literal_eval_memoized(value)
                    except ValueError:
                        pass
            resp[str(issue['issue_jira_id'])].custom_field_values.append(
                JCFVDBData(
                    field_id=jcfv_db_data['field_id'],
                    field_key=jcfv_db_data['field_key'],
                    field_type=jcfv_db_data['field_type'],
                    value=value,
                )
            )
    return resp


@jelly_trace
def _retrieve_custom_fields_from_jellyfish(
    ingest_config: IngestionConfig,
    output_queue: queue.Queue,
    max_issues_to_process: Optional[int] = None,
    limit: int = JELLYFISH_CUSTOM_FIELDS_API_LIMIT,
) -> None:
    """
    Retrieve all custom fields from Jellyfish. Returns dict[str, IssueJCFVDBData] -
    where `str` is the jira_issue_id - to the output queue. Should be spawned in a separate thread.

    Note: 1K is the API hard limit.
    """
    if max_issues_to_process == 0:
        return

    offset = 0
    limit = min((limit, max_issues_to_process)) if max_issues_to_process else limit

    while True:
        base_url = ingest_config.jellyfish_api_base
        headers = {"Jellyfish-API-Token": ingest_config.jellyfish_api_token}
        st = time.perf_counter()

        def _get_values_from_jellyfish() -> requests.Response:
            """define this as a function, so we can use with retry_for_status"""
            try:
                response = requests.get(
                    f"{base_url}/endpoints/jira/issues/custom-fields?offset={offset}&limit={limit}",
                    headers=headers,
                    timeout=JELLYFISH_API_TIMEOUT,
                )
                response.raise_for_status()
            except Exception as e:
                logging_helper.send_to_agent_log_file(
                    f'Error when attempting to pull custom fields from Jellyfish',
                    exc_info=True,
                    level=logging.ERROR,
                )
                raise e
            return response

        # Sometimes we'll see 504's from Jellyfish (e.g. if pulling a large amount of data)
        # so we'll retry for status here.
        r = retry_for_status(_get_values_from_jellyfish)

        data = r.json()
        if not data['issues']:
            logging_helper.send_to_agent_log_file(
                f'No more issues found when attempting to pull custom fields from Jellyfish. {offset} issues retrieved.',
            )
            break

        # API is hard-limited to 10K, which we should be respecting
        # here, but let's do this just-in-case.
        limit = min((limit, data['max_records']))
        # Handle the case where we have a max_issues_to_process, that
        # isn't a clean multiple of limit
        limit = min((limit, max_issues_to_process - offset)) if max_issues_to_process else limit
        offset += data['total_records']

        # Annotate data to be put into the output queue.
        output_queue.put(_annotate_results_from_jellyfish(data))

        if max_issues_to_process is not None and offset >= max_issues_to_process:
            logging_helper.send_to_agent_log_file(
                f'Finished pulling custom fields from Jellyfish - reached max_issues_to_process. {offset} issues retrieved.'
            )
            break
        if data['total_records'] < limit:
            logging_helper.send_to_agent_log_file(
                f'Finished pulling custom fields from Jellyfish - reached end of data. {offset} issues retrieved.'
            )
            break
        logging_helper.send_to_agent_log_file(
            f'Pulled {offset} issues from Jellyfish, {len(data["issues"])} records in {time.perf_counter() - st:.2f}s,'
            f' continuing...',
            level=logging.DEBUG,
        )


class JCFVUpdateFullPayload(NamedTuple):
    missing_from_db_jcfv: list[JCFVUpdate]
    missing_from_jira_jcfv: list[JCFVUpdate]
    out_of_sync_jcfv: list[JCFVUpdate]


@jelly_trace
def identify_custom_field_mismatches(
    ingest_config: IngestionConfig,
    nthreads: int = 10,
    max_issues_to_process: Optional[int] = None,
) -> JCFVUpdateFullPayload:
    """
    At a high level:
    - Get all of the fields that are used to define critical components in Jellyfish. We use a Jellyfish
        endpoint to get this data, in order to support using this sync path from both the agent and internally.
    - Create one thread that retrieves (issue.id, issue.jira_id, *custom_field_values) of all issues
        and puts that into a queue. This is done in a separate thread to allow us to download issues from Jira
        in parallel, as downloading issues from Jira is the primary bottleneck.
    - In our main thread, read from the queue that's being stuffed by the _retrieve_custom_fields_from_jellyfish
        thread, and query Jira to get up-to-date issue data. If there is any data mismatch in custom
        fields, save that into an "updates" NamedTuple, which we can POST to Jellyfish to update the data.

    :param ingest_config: IngestionConfig object
    :param nthreads: Number of threads to use for downloading issues from Jira
    :param max_issues_to_process: Maximum number of issues to process. If None, all issues will be processed.
    :return: Tuple of (missing_from_db_jcfv, missing_from_jira_jcfv, out_of_sync_jcfv)
    """
    output_queue = queue.Queue()

    # Use ThreadPoolExecutor here so that we can capture
    # any exceptions raised within the thread.
    executor = ThreadPoolExecutorWithLogging(max_workers=1)
    retrieval_thread = executor.submit(
        _retrieve_custom_fields_from_jellyfish,
        ingest_config,
        output_queue,
        max_issues_to_process=max_issues_to_process,
    )

    # We _are_ downloading issues but not doing a full download, so for_download=False
    # is fine here and is faster.
    jira_connection = get_jira_connection(config=ingest_config.jira_config)

    # We store the custom field values that are missing from the DB, missing from Jira,
    # and out of sync in separate lists - this allows for easier processing on the
    # update side.
    update_payload = JCFVUpdateFullPayload([], [], [])

    st = time.perf_counter()
    total_issues_scanned = 0
    while retrieval_thread.running() or not output_queue.empty():
        try:
            db_issue_batch: dict[str, IssueJCFVDBData] = output_queue.get(timeout=60.0)
        except queue.Empty:
            logging_helper.send_to_agent_log_file(
                'Didn\'t get any issues from the queue in 60 seconds, retrying...'
            )
            continue

        total_issues_scanned += len(db_issue_batch)

        # Get all the field keys from the DB data, so we can request them from Jira.
        field_map = {}
        for issue in db_issue_batch.values():
            for jcfv in issue.custom_field_values:
                field_map[jcfv.field_key] = (jcfv.field_id, jcfv.field_key, jcfv.field_type)

        # If we don't have any fields to query, we can skip the Jira query.
        if not field_map:
            continue

        # Likely won't change the batch size vs. doing this without "fields", but
        # it's possible we'll be able to use a larger batch size if querying a subset
        # of fields.
        batch_size = get_jira_search_batch_size(
            jira_connection, fields=['key'] + list(field_map.keys())
        )

        # Note: this is parallelized and returns a generator, so we're
        # able to process issues while downloading from Jira simultaneously.
        with record_span('jira_issues_download'):
            jira_issue_batch = pull_jira_issues_by_jira_ids(
                jira_connection=jira_connection,
                jira_ids=list(db_issue_batch.keys()),
                num_parallel_threads=nthreads,
                batch_size=batch_size,
                include_fields=['key'] + list(field_map.keys()),
            )

        # Now we need to compare the custom field values from the DB with the actual issue data from Jira.
        tracer = trace.get_tracer(__name__)
        compare_issues_to_db_span = tracer.start_span(name='compare_issues_to_db')
        for issue_jira in jira_issue_batch:
            issue_db = db_issue_batch[str(issue_jira['id'])]
            jcfv_db_dict = {jcfv.field_key: jcfv for jcfv in issue_db.custom_field_values}

            for field_id, field_key, field_type in field_map.values():
                # Attempt to get the custom field value from the DB.
                # This may not be present, in the case that the field was
                # added to the issue, but we haven't yet pulled the data.
                db_value = getattr(jcfv_db_dict.get(field_key), 'value', None)

                # Occasionally, an issue may not have any of the fields we
                # requested, (e.g. if we're looking for just one field which is
                # no longer present on the issue present) so we need to handle this case.
                jira_value = issue_jira.get('fields', {}).get(field_key, None)

                if db_value != jira_value:
                    if db_value is None:
                        list_to_append = update_payload.missing_from_db_jcfv
                    elif jira_value is None:
                        list_to_append = update_payload.missing_from_jira_jcfv
                    else:
                        list_to_append = update_payload.out_of_sync_jcfv
                    list_to_append.append(
                        JCFVUpdate(
                            field_id=field_id,
                            field_key=field_key,
                            field_type=field_type,
                            jira_issue_id=issue_db.issue_jira_id,
                            value_old=db_value,
                            value_new=jira_value,
                        )
                    )
        compare_issues_to_db_span.end()

        total_fields_out_of_sync = sum(map(len, update_payload))
        logging_helper.send_to_agent_log_file(
            f'Downloaded {total_issues_scanned} issues from Jira. '
            f'{total_fields_out_of_sync} out of sync field values found '
            f'so far in {time.perf_counter() - st} seconds.'
        )
        st = time.perf_counter()

    # Run here in case we get no issues back from Jellyfish
    total_fields_out_of_sync = sum(map(len, update_payload))
    logging_helper.send_to_agent_log_file(
        f'Finished processing all issues from the DB. {total_fields_out_of_sync} '
        f'out of sync field values found in total.'
    )
    logging_helper.send_to_agent_log_file(
        f'{len(update_payload.out_of_sync_jcfv)} custom field values to update.'
    )
    logging_helper.send_to_agent_log_file(
        f'{len(update_payload.missing_from_jira_jcfv)} custom field values to delete.'
    )
    logging_helper.send_to_agent_log_file(
        f'{len(update_payload.missing_from_db_jcfv)} custom field values to insert.'
    )
    logging_helper.send_to_agent_log_file(f'{total_issues_scanned} issues scanned in total.')

    # Retrieval thread is already finished, as guaranteed by the while loop above.
    # We can safely call result() here, which will raise any exceptions we might have seen.
    retrieval_thread.result()

    return update_payload



import json

from jf_ingest.config import AzureDevopsAuthConfig, GitConfig, GitProvider
from jf_ingest.jf_git.adapters.azure_devops import AzureDevopsAdapter
from jf_ingest.jf_git.standardized_models import StandardizedOrganization, StandardizedRepository

PATH_TO_TEST_FIXTURES = 'tests/jf_git/fixtures/ado'
TEST_BASE_URL = 'https://ado.com/'
ADO_ORGANIZATION_LOGIN = 'ADO-Organization'
ADO_PROJECT_NAME = 'ADO-Project-Name'

BASIC_ORG = StandardizedOrganization(
    id=ADO_ORGANIZATION_LOGIN,
    name=ADO_ORGANIZATION_LOGIN,
    login=ADO_ORGANIZATION_LOGIN,
    url=TEST_BASE_URL,
)

BASIC_REPO = StandardizedRepository(
    id=f'STANDARD_REPO_ID',
    name='STANDARD_REPO_NAME',
    full_name=f'STANDARD_REPO_FULL_NAME',
    url=f'{TEST_BASE_URL}{ADO_ORGANIZATION_LOGIN}/{ADO_PROJECT_NAME}/_git/test',
    is_fork=False,
    default_branch_name='master',
    default_branch_sha=None,
    organization=BASIC_ORG,
    commits_backpopulated_to=None,
    prs_backpopulated_to=None,
)


def get_fixture_data(file_name: str):
    with open(file=f'{PATH_TO_TEST_FIXTURES}/{file_name}', mode='r') as f:
        return json.loads(f.read())['value']


def get_raw_users():
    # This raw file is populated from this data set from the ADO docs:
    # https://learn.microsoft.com/en-us/rest/api/azure/devops/graph/users/list?view=azure-devops-rest-7.1&tabs=HTTP
    return get_fixture_data('raw_graph_users.json')


def get_raw_teams():
    # https://learn.microsoft.com/en-us/rest/api/azure/devops/graph/groups/list?view=azure-devops-rest-7.1&tabs=HTTP
    return get_fixture_data('raw_graph_teams.json')


def get_raw_repos():
    # https://learn.microsoft.com/en-us/rest/api/azure/devops/git/repositories/list?view=azure-devops-rest-7.1&tabs=HTTP
    return get_fixture_data('raw_repos.json')


def get_raw_commits():
    return get_fixture_data('raw_commits.json')


def get_raw_branches():
    return get_fixture_data('raw_branches.json')


def get_raw_iterations():
    # I just made this one up, ADO doesn't provide examples
    return get_fixture_data('raw_iterations.json')


def get_raw_prs():
    # https://learn.microsoft.com/en-us/rest/api/azure/devops/git/pull-requests/get-pull-requests?view=azure-devops-rest-7.1&tabs=HTTP
    return get_fixture_data('raw_prs.json')


def get_raw_diffs():
    # https://learn.microsoft.com/en-us/rest/api/azure/devops/git/diffs/get?view=azure-devops-rest-7.1&tabs=HTTP
    with open(file=f'{PATH_TO_TEST_FIXTURES}/raw_diff.json', mode='r') as f:
        return json.loads(f.read())['changes']


def get_raw_change_counts():
    # https://learn.microsoft.com/en-us/rest/api/azure/devops/git/diffs/get?view=azure-devops-rest-7.1&tabs=HTTP
    with open(file=f'{PATH_TO_TEST_FIXTURES}/raw_diff.json', mode='r') as f:
        return json.loads(f.read())['changeCounts']


def get_raw_threads_comments():
    # https://learn.microsoft.com/en-us/rest/api/azure/devops/git/pull-request-threads/list?view=azure-devops-rest-7.1&tabs=HTTP
    return get_fixture_data('raw_threads.json')


def get_adapter(redact_names_and_urls: bool = False, strip_text_content: bool = False):
    auth_config = AzureDevopsAuthConfig(
        company_slug='company_slug',
        token='let-me-in',
        base_url=TEST_BASE_URL,
    )
    config = GitConfig(
        company_slug='test_company',
        instance_slug='test_company_instance_slug',
        instance_file_key='FILE_KEY',
        git_auth_config=auth_config,
        git_organizations=[ADO_ORGANIZATION_LOGIN],
        git_redact_names_and_urls=redact_names_and_urls,
        git_strip_text_content=strip_text_content,
        git_provider=GitProvider.ADO,
    )
    adapter = AzureDevopsAdapter(config)
    adapter.repo_id_to_project_id[BASIC_REPO.id] = ADO_PROJECT_NAME
    return adapter



import json
import logging
import traceback
from collections import defaultdict
from dataclasses import asdict, dataclass, field, fields
from typing import Union

from jira import JIRA, JIRAError
from requests.exceptions import RequestException

from jf_ingest import logging_helper
from jf_ingest.config import (
    GitConfig,
    GitProvider,
    IngestionType,
    JiraAuthConfig,
    JiraDownloadConfig,
)
from jf_ingest.jf_git.adapters import GitAdapter
from jf_ingest.jf_git.exceptions import GitProviderUnavailable
from jf_ingest.jf_git.standardized_models import StandardizedRepository
from jf_ingest.jf_jira.auth import JiraAuthenticationException, JiraAuthMethod, get_jira_connection
from jf_ingest.jf_jira.downloaders import _get_project_filters, download_users
from jf_ingest.utils import (
    PROJECT_HTTP_CODES_TO_RETRY_ON,
    retry_for_status,
    test_jira_or_git_object_access,
)

# NOTE:
# Logger.info will log to stdout AND to our log file.
# We do NOT want to log any passwords or usernames.
# To be extra safe, use print instead of logger.log within validation if you think data could be sensitive.
logger = logging.getLogger(__name__)


@dataclass
class ProjectAccessResult:
    """Jira Specific"""

    issues: bool = False
    versions: bool = False
    components: bool = False


@dataclass
class BoardAccessResult:
    """Jira Specific"""

    sprint_access_ok: bool = False
    sprints: list[str] = field(default_factory=list)


@dataclass
class ObjectAccessResult:
    """General access result for Jira and Git"""

    access_ok: bool = False
    available_objects: list[str] = field(default_factory=list)


@dataclass
class RepositoryAccessResult:
    """Git Specific"""

    repository_name: str
    repository_id: int
    pull_requests: bool = False
    branches: bool = False
    commits_on_default_branch: bool = False


def _bar_factory():
    """
    Annoyingly, dataclasses.field required a callable that takes no args.
    This satisfies that requirement.
    """
    return defaultdict(BoardAccessResult)


def _par_factory():
    """
    Annoyingly, dataclasses.field required a callable that takes no args.
    This satisfies that requirement.
    """
    return defaultdict(ProjectAccessResult)


class SafeDumpDataClass:
    """
    dataclasses.asdict does not work when the values of the dataclasses can be default dicts.
    Let's fix that.

    On version 3.12 we can remove this.
    https://github.com/python/cpython/pull/32056
    https://docs.python.org/3/whatsnew/changelog.html#python-3-12-0-alpha-1
    """

    @classmethod
    def from_dict(cls, body: dict):
        return cls(**body)

    def to_dict(self):
        # convert defaultdict fields to plain old dicts so we can dump the data
        _fields = fields(self)
        for field in _fields:
            value = getattr(self, field.name)
            if isinstance(value, defaultdict):
                setattr(self, field.name, dict(value))
            # recursively handle the sub-healthcheck results on IngestionHealthCheckResult
            elif isinstance(
                value, (GitConnectionHealthCheckResult, JiraConnectionHealthCheckResult)
            ):
                setattr(self, field.name, value.to_dict())
        return asdict(self)


@dataclass
class JiraConnectionHealthCheckResult(SafeDumpDataClass):
    """
    Representing the result of a Jira connection healthcheck report.
    """

    successful: bool

    # The version of Jira that is being run
    server_version: str = None

    # Total list of projects accessible
    accessible_projects: list[str] = field(default_factory=list)

    # Projects included in the ingestion configuration to be downloaded but inaccessible from the validator.
    included_inaccessible_projects: list[str] = field(default_factory=list)

    # Access to project-specific data
    project_data: dict = field(default_factory=_par_factory)

    # Fields available on the instance
    fields: ObjectAccessResult = field(default_factory=ObjectAccessResult)

    # Included fields that we cannot access
    included_inaccessible_fields: list[str] = field(default_factory=list)

    # Resolutions available on the instance
    resolutions: ObjectAccessResult = field(default_factory=ObjectAccessResult)

    # Issue Types available on the instance
    issue_types: ObjectAccessResult = field(default_factory=ObjectAccessResult)

    # Issue Link Types available on the instance
    issue_link_types: ObjectAccessResult = field(default_factory=ObjectAccessResult)

    # Priorities available on the instance
    priorities: ObjectAccessResult = field(default_factory=ObjectAccessResult)

    # Boards available on the instance
    boards: ObjectAccessResult = field(default_factory=ObjectAccessResult)

    # Sprints available on the instance by Board
    boards_and_sprints: dict = field(default_factory=_bar_factory)

    # How many users the validator can access
    num_accessible_users: int = 0

    # Permissions granted to the user.
    granted_permissions: list[str] = field(default_factory=list)

    # Whether we have been able to receive an X-ANODE-ID header from the Jira connection.
    # This means the customer is a Jira Data Center customer.
    returns_anode_id_header: bool = False


@dataclass
class GitConnectionHealthCheckResult(SafeDumpDataClass):
    """
    Representing the result of a Git connection healthcheck report, broken down by Git Organization
    """

    instance_slug: str
    organization_login: str

    successful: bool

    git_provider: GitProvider

    api_scopes: str = None

    # users available on the instance
    users: ObjectAccessResult = field(default_factory=ObjectAccessResult)

    # teams available on the instance
    teams: ObjectAccessResult = field(default_factory=ObjectAccessResult)

    # Dict of organization name : repo list of repos we can access
    accessible_repos: list[str] = field(default_factory=list)
    repo_data: dict[str, RepositoryAccessResult] = field(default_factory=dict)

    # list of repos explicitly included in the config but inaccessible with our credentials
    # This is specific only to the Agent
    included_inaccessible_repos: list[str] = field(default_factory=list)

    @classmethod
    def from_dict(cls, body: dict):
        return cls(**body)


@dataclass
class IngestionHealthCheckResult(SafeDumpDataClass):
    """
    Dataclass representing the result of a pre-ingestion permission healthcheck run.
    """

    ingestion_type: IngestionType

    # Whether the healthcheck result was fully successful
    fully_successful: bool = False

    # Customers may have multiple git configs, so we want to account for that with multiple healthcheck results.
    # The key here should be the instance slug.
    git_connection_healthcheck: list[GitConnectionHealthCheckResult] = field(default_factory=list)

    jira_connection_healthcheck: JiraConnectionHealthCheckResult = None

    def __post_init__(self):
        """
        Sets the fully_successful field based on the successful fields on the jira and git inputs.
        """
        jira_successful = (
            self.jira_connection_healthcheck.successful
            if self.jira_connection_healthcheck
            else False
        )

        # If we have a non-empty git healthcheck result list, ensure all its members are successful.
        git_successful = bool(self.git_connection_healthcheck) and all(
            [x.successful for x in self.git_connection_healthcheck]
        )

        self.fully_successful = jira_successful and git_successful

    def to_json(self) -> str:
        """
        Converts to a JSON string.

        Returns: The IngestionHealthCheckResult as a JSON representation.

        """
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, body: str):
        """
        Deserializes a JSON string into an IngestionHealthCheckResult

        Args:
            body:

        Returns: The IngestionHealthCheckResult as a properly nested dataclass

        """
        healthcheck_dict = json.loads(body)

        # We need to do this because Python dataclasses won't natively deserialize a dict representation of a nested
        # dataclass (without third-party packages), so we intercept the list[dict] result
        # and convert it into a list[dataclass].

        new_git_healthcheck_list = []

        for healthcheck_result in healthcheck_dict["git_connection_healthcheck"]:
            parsed_result = GitConnectionHealthCheckResult.from_dict(healthcheck_result)
            new_git_healthcheck_list.append(parsed_result)

        healthcheck_dict["git_connection_healthcheck"] = new_git_healthcheck_list

        healthcheck_dict["jira_connection_healthcheck"] = (
            JiraConnectionHealthCheckResult.from_dict(
                healthcheck_dict["jira_connection_healthcheck"]
            )
            if healthcheck_dict["jira_connection_healthcheck"]
            else None
        )

        return cls(**healthcheck_dict)


def _print_debug_jira_error(e: JIRAError, config: JiraDownloadConfig):
    print("Response:")
    print("  Headers:", e.headers)
    print("  URL:", e.url)
    print("  Status Code:", e.status_code)
    print("  Text:", e.text)

    if "Basic authentication with passwords is deprecated." in str(e):
        logger.error(
            f"Error connecting to Jira instance at {config.url}. Please use a Jira API token, see https://confluence.atlassian.com/cloud/api-tokens-938839638.html"
        )
    else:
        logger.error(
            f"Error connecting to Jira instance at {config.url}, please validate your credentials. Error: {e}"
        )


def _attempt_anode_id_header_check(jira_connection: JIRA) -> bool:
    """

    Args:
        jira_connection: The Jira Connection to query over

    Returns: Whether the X-ANODEID header exists on the response.
    If it does, this means the customer is Jira Data Center.

    """
    try:
        logger.info("==> Getting Jira deployment type...")

        res = jira_connection._session.get(jira_connection._get_url("serverInfo"))
        headers = res.headers

        if "X-ANODEID" in headers:
            logger.info(
                "Response headers contains X-ANODEID! Customer is running Jira Data Center."
            )

            return True
        else:
            logger.info(
                "Response headers does not contain X-ANODEID! Customer is NOT running Jira Data Center."
            )

            return False

    except Exception:
        # This is a pretty hacky check using private members of the jira connection class
        # So we can't rely on it always working. If it breaks we should just return False
        # and move on.
        logger.error("Unable to get X-ANODEID headers from request!")

        return False


def get_jira_version(jira_connection: JIRA) -> str | None:
    """

    Args:
        jira_connection: The Jira connection

    Returns: the string representation of the Jira server version.

    """

    logger.info("==> Getting Jira version...")

    server_info = jira_connection.server_info()

    jira_version = server_info["version"] if "version" in server_info.keys() else None

    logger.info(f"Found Jira version as {jira_version}")

    return jira_version


def get_jira_permissions(jira_connection: JIRA) -> list[str]:
    """
    Gets the Jira permissions we know about for this user (experimental)
    Args:
        jira_connection: The jira connection to request over.

    Returns: A list of permissions that the user has.

    """

    logger.info("==> Getting Jira permissions...")

    # These are most of the permissions that Jira says are built-in, per this doc:
    # https://developer.atlassian.com/cloud/jira/platform/rest/v2/api-group-permission-schemes/#built-in-permissions
    # We can pare this down if we decide we don't care about any of these.
    permission_set = {
        'ADD_COMMENTS',
        'ADMINISTER_PROJECTS',
        'ASSIGNABLE_USER',
        'ASSIGN_ISSUES',
        'BROWSE_PROJECTS',  # required to use JF. https://help.jellyfish.co/hc/en-us/articles/13428996347277-Connecting-Jira-to-Jellyfish
        'CLOSE_ISSUES',
        'CREATE_ISSUES',
        'DELETE_ALL_COMMENTS',
        'DELETE_ALL_WORKLOGS',
        'DELETE_ISSUES',
        'DELETE_OWN_COMMENTS',
        'DELETE_OWN_WORKLOGS',
        'EDIT_ALL_COMMENTS',
        'EDIT_ALL_WORKLOGS',
        'EDIT_ISSUES',
        'EDIT_OWN_COMMENTS',
        'EDIT_OWN_WORKLOGS',
        'LINK_ISSUES',
        'MODIFY_REPORTER',
        'MOVE_ISSUES',
        'RESOLVE_ISSUES',
        'SCHEDULE_ISSUES',
        'SET_ISSUE_SECURITY',
        'TRANSITION_ISSUES',
        'USER_PICKER',  # required to use JF. https://help.jellyfish.co/hc/en-us/articles/13428996347277-Connecting-Jira-to-Jellyfish
        'VIEW_READONLY_WORKFLOW',
        'WORK_ON_ISSUES',
    }

    params = {"permissions": ", ".join(list(permission_set))}

    # Jira has some changes here, and we want to be backwards and forwards compatible.
    # Our version of the Jira client does *not* support this parameter right now, so we manually do it.
    # If we update the Jira version we can instead do jira_connection.my_permissions
    # https://developer.atlassian.com/cloud/jira/platform/change-notice-get-my-permissions-requires-permissions-query-parameter/#let-s-see-the-code

    permissions = jira_connection._get_json("mypermissions", params=params)

    granted_permissions = []

    for key, value in permissions["permissions"].items():
        if value["havePermission"]:
            granted_permissions.append(key)

    logger.info(f"Found granted permissions as {granted_permissions}")

    return granted_permissions


def validate_jira(
    config: Union[JiraDownloadConfig, JiraAuthConfig]
) -> JiraConnectionHealthCheckResult:
    """
    Validates jira configuration and credentials. Returns a JiraHealthcheckResult object
    representing whether the check was successful and what the errors were, if any.
    Modified from the original Jira validation logic in the Agent.
    """

    healthcheck_result = JiraConnectionHealthCheckResult(successful=True)

    print("\nJira details:")
    print(f"  URL:      {config.url}")
    print(f"  Username: {config.user}")

    if config.user and config.password:
        print("  Password: **********")
    elif config.personal_access_token:
        print("  Token: **********")
    else:
        logger.error("No Jira credentials found in Jira authentication config!")
        healthcheck_result.successful = False

        return healthcheck_result

    # test Jira connection
    try:
        logger.info("==> Testing Jira connection...")
        jira_connection = get_jira_connection(
            config=config, auth_method=JiraAuthMethod.BasicAuth, max_retries=1
        )
        jira_connection.myself()

        healthcheck_result.server_version = get_jira_version(jira_connection)
        healthcheck_result.returns_anode_id_header = _attempt_anode_id_header_check(jira_connection)
        healthcheck_result.granted_permissions = get_jira_permissions(jira_connection)

    except (JIRAError, JiraAuthenticationException) as e:
        print(e)
        # we want to grab the underlying exception on JiraAuthenticationException
        # so we can provide the maximum amount of useful data to someone running in validate mode.
        # if we don't have a JIRAError as the current or underlying exception, call it quits after
        # printing above.
        if type(e) == JiraAuthenticationException and type(e.original_exception) == JIRAError:
            _print_debug_jira_error(e.original_exception, config)
        elif type(e) == JIRAError:
            _print_debug_jira_error(e, config)

        healthcheck_result.successful = False

        return healthcheck_result

    except RequestException as e:
        logger.error(f"RequestException when validating Jira! Message: {e}")

        # Print debugging information related to the request exception
        if e.request:
            print("Request:")
            print("  URL:", e.request.method, e.request.url)
            print("  Body:", e.request.body)
        else:
            print('RequestException contained no "request" value.')

        if e.response:
            print("Response:")
            print("  Headers:", e.response.headers)
            print("  URL:", e.response.url)
            print("  Status Code:", e.response.status_code)
            print("  Text:", e.response.text)
        else:
            print('RequestException contained no "response" value.')

        healthcheck_result.successful = False

        return healthcheck_result

    except Exception as e:
        raise

    # test jira users permission
    try:
        logger.info("==> Testing Jira user browsing permissions...")
        user_count = len(
            download_users(
                jira_basic_connection=jira_connection,
                jira_atlas_connect_connection=None,
                gdpr_active=config.gdpr_active,
            )
        )
        logger.info(f"We can access {user_count} Jira users.")

        healthcheck_result.num_accessible_users = user_count

    except Exception as e:
        logger.error(
            f'Error downloading users from Jira instance at {config.url}, please verify that this user has the "browse all users" permission. Error: {e}'
        )
        healthcheck_result.successful = False

    # test jira project access
    # Calling connection.projects() will return all project keys to us, even if they are private
    # and we don't have access to them. Wait to add projects to the accessible list until we are
    # certain we can get what we need.
    logger.info("==> Testing Jira project permissions...")

    filters: list = _get_project_filters(
        include_projects=config.include_projects,
        exclude_projects=config.exclude_projects,
        include_categories=config.include_project_categories,
        exclude_categories=config.exclude_project_categories,
    )

    all_projects = retry_for_status(
        jira_connection.projects, statuses_to_retry=PROJECT_HTTP_CODES_TO_RETRY_ON
    )
    projects = {p.key for p in all_projects if all(filt(p) for filt in filters)}

    logger.info(
        f"With provided credentials, the following projects are discoverable: {projects}.\nChecking project access."
    )

    accessible_projects, inaccessible_projects = [], []
    for project_key in sorted(projects):
        logger.info(f'Testing access for project: "{project_key}"')
        healthcheck_result.project_data[project_key].issues, _ = test_jira_or_git_object_access(
            jira_connection.search_issues,
            f'project="{project_key}"',
            fields=['id'],
            maxResults=1,
            return_objs=False,
        )
        healthcheck_result.project_data[project_key].versions, _ = test_jira_or_git_object_access(
            jira_connection.project_versions, project_key, return_objs=False
        )
        healthcheck_result.project_data[project_key].components, _ = test_jira_or_git_object_access(
            jira_connection.project_components, project_key, return_objs=False
        )

        all_good = all(healthcheck_result.project_data[project_key].__dict__.values())

        if all_good:
            accessible_projects.append(project_key)
            logger.info(
                f'With provided credentials, we can access issues, versions, and components within project {project_key}'
            )
        else:
            logger.warning(
                f'Could not access all required data in project {project_key}. Access report:\n'
                f'    issues:     {healthcheck_result.project_data[project_key].issues}\n'
                f'    versions:   {healthcheck_result.project_data[project_key].versions}\n'
                f'    components: {healthcheck_result.project_data[project_key].components}\n'
            )

    if config.include_projects:
        for proj in config.include_projects:
            if proj not in accessible_projects:
                inaccessible_projects.append(proj)

    healthcheck_result.accessible_projects = accessible_projects
    healthcheck_result.included_inaccessible_projects = inaccessible_projects

    if inaccessible_projects:
        project_list_str = ", ".join(inaccessible_projects)
        logger.warning(f"\nUnable to access the following projects: {project_list_str}.")
        if config.include_projects and not set(config.include_projects).issubset(
            accessible_projects
        ):
            logger.error(
                "Unable to access the following projects explicitly provided in the config file: "
                ", ".join([p for p in config.include_projects if p not in accessible_projects])
            )
            healthcheck_result.successful = False

    logger.info('Checking access to fields')
    (
        healthcheck_result.fields.access_ok,
        healthcheck_result.fields.available_objects,
    ) = test_jira_or_git_object_access(jira_connection.fields)
    if config.include_fields:
        healthcheck_result.included_inaccessible_fields = [
            f
            for f in config.include_fields
            not in [field.name for field in retry_for_status(jira_connection.fields)]
        ]
    if healthcheck_result.included_inaccessible_fields:
        logger.error(
            "Unable to access the following fields that were explicitly included in the config file: "
            ", ".join(healthcheck_result.included_inaccessible_fields)
        )

    logger.info('Checking access to resolutions')
    (
        healthcheck_result.resolutions.access_ok,
        healthcheck_result.resolutions.available_objects,
    ) = test_jira_or_git_object_access(jira_connection.resolutions)

    logger.info('Checking access to issue types')
    (
        healthcheck_result.issue_types.access_ok,
        healthcheck_result.issue_types.available_objects,
    ) = test_jira_or_git_object_access(jira_connection.issue_types)

    logger.info('Checking access to issue link types')
    (
        healthcheck_result.issue_link_types.access_ok,
        healthcheck_result.issue_link_types.available_objects,
    ) = test_jira_or_git_object_access(jira_connection.issue_link_types)

    logger.info('Checking access to priorities')
    (
        healthcheck_result.priorities.access_ok,
        healthcheck_result.priorities.available_objects,
    ) = test_jira_or_git_object_access(jira_connection.priorities)

    logger.info('Checking access to boards')
    (
        healthcheck_result.boards.access_ok,
        healthcheck_result.boards.available_objects,
    ) = test_jira_or_git_object_access(jira_connection.boards)

    if healthcheck_result.boards.access_ok:
        boards = retry_for_status(jira_connection.boards)
        logger.info('Checking access to sprints')
        for board in boards:
            (
                healthcheck_result.boards_and_sprints[board.name].sprint_access_ok,
                healthcheck_result.boards_and_sprints[board.name].sprints,
            ) = test_jira_or_git_object_access(jira_connection.sprints, board.id, is_sprint=True)

    # TODO include repo data here?
    # We currently dont flag repos as inaccessible in the UI, so shouldn't fail the overall
    # check yet if we're not able to have the user dig into the repo data.
    # Probably the best thing to do here is a similar approach to Jira projects, where we
    # only fail if we've explicitly included a repo in the config and can't access it.
    healthcheck_result.successful &= all(
        [
            healthcheck_result.fields.access_ok,
            healthcheck_result.resolutions.access_ok,
            healthcheck_result.issue_types.access_ok,
            healthcheck_result.issue_link_types.access_ok,
            healthcheck_result.priorities.access_ok,
            healthcheck_result.boards.access_ok,
            healthcheck_result.num_accessible_users,
            len(healthcheck_result.accessible_projects),
            not len(healthcheck_result.included_inaccessible_projects),
        ]
        + [bar.sprint_access_ok for bar in healthcheck_result.boards_and_sprints.values()]
    )

    return healthcheck_result


def validate_git(git_configs: list[GitConfig]) -> list[GitConnectionHealthCheckResult]:
    """Attempts to hit all relevant GIT APIs and return a health check verifying our access

    Args:
        git_configs (list[GitConfig]): A list of git configs to crawl over
        get_all_available_objects (bool, optional): A boolean flag that when set to True will query for all available objects. Defaults to False.

    Returns:
        list[GitConnectionHealthCheckResult]: A health check object detailing what we do and do not have access to
    """

    healthcheck_result_list: list[GitConnectionHealthCheckResult] = []

    for i, git_config in enumerate(git_configs, start=1):
        logger.info(f"Git Configuration details for instance {i}/{len(git_configs)}:")
        logger.info(f"  Instance slug: {git_config.instance_slug}")
        logger.info(f"  Provider: {git_config.git_provider.value}")
        logger.info(f"  Organizations: {len(git_config.git_organizations)}")
        if len(git_config.excluded_organizations) > 0:
            logger.info(f"  Excluded organizations: {git_config.excluded_organizations}")
        if len(git_config.included_repos) > 0:
            logger.info(f"  Included repos: {git_config.included_repos}")
        if len(git_config.excluded_repos) > 0:
            logger.info(f"  Excluded repos: {git_config.excluded_repos}")
        if len(git_config.included_branches_by_repo) > 0:
            logger.info(f"  Included Branches: {git_config.included_branches_by_repo}")

        logger.info('==> Testing Git connection...')

        try:
            git_adapter = GitAdapter.get_git_adapter(git_config)
        except Exception as e:
            logger.error(
                f'Error connecting to Git instance {git_config.instance_slug}. Exception: {e}\n'
            )
            logging_helper.send_to_agent_log_file(traceback.format_exc(), level=logging.ERROR)

            healthcheck_result = GitConnectionHealthCheckResult(
                successful=False,
                instance_slug=git_config.instance_slug,
                organization_login=None,
                git_provider=git_config.git_provider.value,
            )
            healthcheck_result_list.append(healthcheck_result)
            continue

        for org in sorted(git_adapter.get_organizations(), key=lambda org: org.login):
            successful = True
            repos = []
            api_scopes = git_adapter.get_api_scopes()
            healthcheck_result = GitConnectionHealthCheckResult(
                successful=successful,
                instance_slug=git_config.instance_slug,
                organization_login=org.login,
                api_scopes=api_scopes,
                git_provider=git_config.git_provider.value,
            )
            if git_adapter.client.uses_jwt:
                logger.info('Using JWT token for authentication. No API Scopes available.')
            else:
                logger.info(f'\nFor Organization {org.name} we have the following API Scopes:')
                logger.info(f'  {api_scopes}\n')

            try:
                (
                    healthcheck_result.users.access_ok,
                    healthcheck_result.users.available_objects,
                ) = test_jira_or_git_object_access(git_adapter.get_users, org, return_attr='login')

                (
                    healthcheck_result.teams.access_ok,
                    healthcheck_result.teams.available_objects,
                ) = test_jira_or_git_object_access(git_adapter.get_teams, org)

                repos: list[StandardizedRepository] = git_adapter.get_repos(org)
                for repo in sorted(repos, key=lambda repo: repo.name):
                    # NOTE: I'm paranoid about if repo names are unique, so I'm combining them with their ID
                    # so we have readability and uniqueness
                    repo_access_result = RepositoryAccessResult(
                        repository_name=repo.name, repository_id=repo.id
                    )
                    repo_access_result.pull_requests = test_jira_or_git_object_access(
                        git_adapter.get_prs,
                        repo,
                        limit=1,
                        return_attr='id',
                    )[0]
                    repo_access_result.branches = bool(repo.branches)
                    repo_access_result.commits_on_default_branch = test_jira_or_git_object_access(
                        git_adapter.get_commits_for_default_branch,
                        repo,
                        limit=1,
                        return_attr='hash',
                    )[0]

                    repo_name_with_id = f'{repo.name}_({repo.id})'
                    healthcheck_result.repo_data[repo_name_with_id] = repo_access_result

                    # Report to the console what we found:
                    if (
                        repo_access_result.pull_requests
                        and repo_access_result.branches
                        and repo_access_result.commits_on_default_branch
                    ):
                        logger.info(
                            f'With provided credentials we can access PRs, Branches, and Commits for Repository {repo_access_result.repository_name} (Git Organization: {org.login})'
                        )
                    else:
                        logger.info(
                            f'Could not access all required data in Repo {repo_access_result.repository_name} (Git Organization: {org.login})'
                        )
                        logger.info(f'    PRs:      {repo_access_result.pull_requests}')
                        logger.info(f'    Commits:  {repo_access_result.commits_on_default_branch}')
                        logger.info(f'    Branches: {repo_access_result.branches}')

            except GitProviderUnavailable as e:
                logger.warning(
                    f'The Git provider is currently not supported in JF Ingest! Provider: {git_config.git_provider}. This config ({git_config.instance_slug}) will be skipped'
                )
                successful = False
            except Exception as e:
                logger.error(f"Git connection unsuccessful! Exception: {e}")
                logger.debug(traceback.format_exc())
                successful = False
            finally:
                healthcheck_result.successful = successful
                healthcheck_result.accessible_repos = [repo.name for repo in repos]
                healthcheck_result.included_inaccessible_repos = [
                    repo_name
                    for repo_name in git_config.included_repos
                    if repo_name not in healthcheck_result.accessible_repos
                ]

        healthcheck_result_list.append(healthcheck_result)

    return healthcheck_result_list



import copy
import json
from contextlib import contextmanager
from datetime import datetime, timezone
from typing import List
from unittest.mock import patch

from dateutil import parser

from jf_ingest.jf_git.adapters import GitAdapter
from jf_ingest.jf_git.adapters.azure_devops import AzureDevopsAdapter
from jf_ingest.jf_git.standardized_models import (
    StandardizedCommit,
    StandardizedTeam,
    StandardizedUser,
)
from jf_ingest.utils import hash_filename
from tests.jf_git._ado_adapter_utils import (
    ADO_PROJECT_NAME,
    BASIC_ORG,
    BASIC_REPO,
    get_adapter,
    get_raw_branches,
    get_raw_change_counts,
    get_raw_commits,
    get_raw_diffs,
    get_raw_iterations,
    get_raw_prs,
    get_raw_repos,
    get_raw_teams,
    get_raw_threads_comments,
    get_raw_users,
)


def test_project_name_from_url():
    adapter = get_adapter()
    project_name = adapter._project_name_from_repo(BASIC_REPO)
    assert project_name == ADO_PROJECT_NAME


def test_get_organizations():
    adapter = get_adapter()
    organizations = adapter.get_organizations()

    assert len(organizations) == 1
    organization = organizations[0]

    assert BASIC_ORG == organization


def test_get_organizations_redact_urls_and_names():
    adapter = get_adapter(redact_names_and_urls=True)
    organizations = adapter.get_organizations()

    org = copy.deepcopy(BASIC_ORG)
    org.name = 'redacted-0000'
    org.url = None

    assert len(organizations) == 1
    organization = organizations[0]

    assert org == organization


def test_get_users():
    adapter = get_adapter()
    raw_users = get_raw_users()
    with patch(
        'jf_ingest.jf_git.clients.azure_devops.AzureDevopsClient.get_graph_users',
        return_value=raw_users,
    ):
        users = list(adapter.get_users(standardized_org=BASIC_ORG))
        raw_user_index = 0
        standard_user_index = 0
        while standard_user_index < len(users):
            raw_user = raw_users[raw_user_index]
            standard_user = users[standard_user_index]
            assert raw_user['descriptor'] == standard_user.id
            assert raw_user['principalName'] == standard_user.login
            assert raw_user['displayName'] == standard_user.name
            assert raw_user['mailAddress'] == standard_user.email
            assert raw_user['url'] == standard_user.url
            raw_user_index += 1
            standard_user_index += 1


def test_get_users_with_redact():
    # NOTE Redacting/stripping variables don't affect user ingestion at all
    adapter = get_adapter(redact_names_and_urls=True, strip_text_content=True)
    raw_users = get_raw_users()
    with patch(
        'jf_ingest.jf_git.clients.azure_devops.AzureDevopsClient.get_graph_users',
        return_value=raw_users,
    ):
        users = list(adapter.get_users(standardized_org=BASIC_ORG))
        raw_user_index = 0
        standard_user_index = 0
        while standard_user_index < len(users):
            raw_user = raw_users[raw_user_index]
            standard_user = users[standard_user_index]
            assert raw_user['descriptor'] == standard_user.id
            assert raw_user['principalName'] == standard_user.login
            assert raw_user['displayName'] == standard_user.name
            assert raw_user['mailAddress'] == standard_user.email
            assert raw_user['url'] == standard_user.url
            raw_user_index += 1
            standard_user_index += 1


def test_get_teams():
    # NOTE Redacting/stripping variables don't affect user ingestion at all
    adapter = get_adapter(redact_names_and_urls=True, strip_text_content=True)
    raw_teams = get_raw_teams()
    raw_users = get_raw_users()
    # TODO: Also patch team members
    with patch(
        'jf_ingest.jf_git.clients.azure_devops.AzureDevopsClient.get_teams', return_value=raw_teams
    ), patch(
        'jf_ingest.jf_git.clients.azure_devops.AzureDevopsClient.get_team_users',
        return_value=raw_users,
    ):
        teams: List[StandardizedTeam] = list(adapter.get_teams(standardized_organization=BASIC_ORG))

        for raw_team, standard_team in zip(raw_teams, teams):
            assert raw_team['descriptor'] == standard_team.id
            assert raw_team['descriptor'] == standard_team.slug
            assert raw_team['displayName'] == standard_team.name
            assert raw_team['description'] == standard_team.description

            members = standard_team.members
            raw_user_index = 0
            standard_user_index = 0
            while standard_user_index < len(members):
                raw_user = raw_users[raw_user_index]
                standard_user = members[standard_user_index]
                assert raw_user['descriptor'] == standard_user.id
                assert raw_user['principalName'] == standard_user.login
                assert raw_user['displayName'] == standard_user.name
                assert raw_user['mailAddress'] == standard_user.email
                assert raw_user['url'] == standard_user.url
                raw_user_index += 1
                standard_user_index += 1


def test_get_repos():
    # NOTE Redacting/stripping variables don't affect user ingestion at all
    adapter = get_adapter(redact_names_and_urls=False, strip_text_content=False)
    raw_repos = get_raw_repos()
    with patch(
        'jf_ingest.jf_git.clients.azure_devops.AzureDevopsClient.get_all_repos',
        return_value=raw_repos,
    ):
        standard_repos = adapter.get_repos(standardized_organization=BASIC_ORG)
        for raw_repo, standard_repo in zip(raw_repos, standard_repos):
            assert raw_repo['id'] == standard_repo.id
            assert f"{BASIC_ORG.login}/{raw_repo['name']}" == standard_repo.full_name
            assert raw_repo['url'] == standard_repo.url
            assert raw_repo.get('isFork', False) == standard_repo.is_fork
            assert BASIC_ORG == standard_repo.organization
            assert (
                raw_repo['defaultBranch'].replace('refs/heads/', '')
                if 'defaultBranch' in raw_repo
                else None
            ) == standard_repo.default_branch_name
            assert None == standard_repo.default_branch_sha


def test_get_repos_with_redaction():
    # NOTE Redacting/stripping variables don't affect user ingestion at all
    adapter = get_adapter(redact_names_and_urls=True, strip_text_content=True)
    raw_repos = get_raw_repos()
    with patch(
        'jf_ingest.jf_git.clients.azure_devops.AzureDevopsClient.get_all_repos',
        return_value=raw_repos,
    ):
        standard_repos = adapter.get_repos(standardized_organization=BASIC_ORG)
        i = 0
        for raw_repo, standard_repo in zip(raw_repos, standard_repos):
            default_branch = (
                raw_repo['defaultBranch'].replace('refs/heads/', '')
                if 'defaultBranch' in raw_repo
                else None
            )
            assert raw_repo['id'] == standard_repo.id
            assert f"{BASIC_ORG.login}/redacted-000{i}" == standard_repo.full_name
            assert None == standard_repo.url
            assert raw_repo.get('isFork', False) == standard_repo.is_fork
            assert BASIC_ORG == standard_repo.organization
            # HACK - quickly swap out branch name if it should be redacted
            if default_branch == 'TEST-BRANCH-REDACT-ME':
                default_branch = 'redacted-0000'
            assert default_branch == standard_repo.default_branch_name
            assert None == standard_repo.default_branch_sha
            # Kinda hacky, but spoof the name redactor by incrementing this variable
            i += 1


def test_get_commits_for_default_branch():
    # NOTE Redacting/stripping variables don't affect user ingestion at all
    adapter = get_adapter(redact_names_and_urls=False, strip_text_content=False)
    raw_commits = get_raw_commits()
    with patch(
        'jf_ingest.jf_git.clients.azure_devops.AzureDevopsClient.get_commits',
        return_value=raw_commits,
    ):
        standard_commits = adapter.get_commits_for_default_branch(standardized_repo=BASIC_REPO)
        for raw_commit, standard_commit in zip(raw_commits, standard_commits):
            assert standard_commit.hash == raw_commit['commitId']
            assert standard_commit.url == raw_commit['url']
            assert standard_commit.message == raw_commit['comment']
            assert standard_commit.branch_name == BASIC_REPO.default_branch_name
            assert standard_commit.commit_date == parser.parse(raw_commit['committer']['date'])
            assert standard_commit.author_date == parser.parse(raw_commit['author']['date'])
            assert type(standard_commit.author) == StandardizedUser
            assert standard_commit.author != None
            assert standard_commit.repo.id == BASIC_REPO.id
            assert standard_commit.repo.url == BASIC_REPO.url
            assert standard_commit.repo.name == BASIC_REPO.name
            assert standard_commit.is_merge == False


def test_get_commits_for_default_branch_with_redaction():
    # NOTE Redacting/stripping variables don't affect user ingestion at all
    adapter = get_adapter(redact_names_and_urls=True, strip_text_content=True)
    raw_commits = get_raw_commits()
    with patch(
        'jf_ingest.jf_git.clients.azure_devops.AzureDevopsClient.get_commits',
        return_value=raw_commits,
    ):
        standard_commits = adapter.get_commits_for_default_branch(standardized_repo=BASIC_REPO)
        for raw_commit, standard_commit in zip(raw_commits, standard_commits):
            assert standard_commit.hash == raw_commit['commitId']
            assert standard_commit.url == None
            assert standard_commit.message == GitAdapter.sanitize_text(raw_commit['comment'], True)
            assert standard_commit.branch_name == BASIC_REPO.default_branch_name
            assert standard_commit.commit_date == parser.parse(raw_commit['committer']['date'])
            assert standard_commit.author_date == parser.parse(raw_commit['author']['date'])
            assert type(standard_commit.author) == StandardizedUser
            assert standard_commit.author != None
            assert standard_commit.repo.id == BASIC_REPO.id
            assert standard_commit.repo.url == BASIC_REPO.url
            assert standard_commit.repo.name == BASIC_REPO.name
            assert standard_commit.is_merge == False


def test_get_commits_for_default_branch_with_only_strip_text():
    # NOTE Redacting/stripping variables don't affect user ingestion at all
    adapter = get_adapter(redact_names_and_urls=False, strip_text_content=True)
    raw_commits = get_raw_commits()
    with patch(
        'jf_ingest.jf_git.clients.azure_devops.AzureDevopsClient.get_commits',
        return_value=raw_commits,
    ):
        standard_commits = adapter.get_commits_for_default_branch(standardized_repo=BASIC_REPO)
        for raw_commit, standard_commit in zip(raw_commits, standard_commits):
            assert standard_commit.hash == raw_commit['commitId']
            assert standard_commit.url == raw_commit['url']
            assert standard_commit.message == GitAdapter.sanitize_text(raw_commit['comment'], True)
            assert standard_commit.branch_name == BASIC_REPO.default_branch_name
            assert standard_commit.commit_date == parser.parse(raw_commit['committer']['date'])
            assert standard_commit.author_date == parser.parse(raw_commit['author']['date'])
            assert type(standard_commit.author) == StandardizedUser
            assert standard_commit.author != None
            assert standard_commit.repo.id == BASIC_REPO.id
            assert standard_commit.repo.url == BASIC_REPO.url
            assert standard_commit.repo.name == BASIC_REPO.name
            assert standard_commit.is_merge == False


def test_get_default_branch_for_repo():
    # NOTE Redacting/stripping variables don't affect user ingestion at all
    adapter = get_adapter(redact_names_and_urls=False, strip_text_content=True)

    branches = list(adapter.get_branches_for_repo(BASIC_REPO, pull_branches=False))
    branch = branches[0]

    assert branch.is_default == True
    assert branch.name == BASIC_REPO.default_branch_name
    assert branch.repo_id == BASIC_REPO.id
    assert branch.sha == BASIC_REPO.default_branch_sha


def test_get_branches_for_repo():
    # NOTE Redacting/stripping variables don't affect user ingestion at all
    adapter = get_adapter(redact_names_and_urls=False, strip_text_content=False)
    raw_branches = get_raw_branches()

    with patch(
        'jf_ingest.jf_git.clients.azure_devops.AzureDevopsClient.get_branches',
        return_value=raw_branches,
    ):
        standard_branches = adapter.get_branches_for_repo(BASIC_REPO, pull_branches=True)
        for standard_branch, raw_branch in zip(standard_branches, raw_branches):
            branch_name = raw_branch['name'].replace('refs/heads/', '')
            assert standard_branch.is_default == (
                standard_branch.name == BASIC_REPO.default_branch_name
            )
            assert standard_branch.name == branch_name
            assert standard_branch.repo_id == BASIC_REPO.id
            assert standard_branch.sha == raw_branch['objectId']


def test_get_branches_for_repo_with_redaction():
    # NOTE Redacting/stripping variables don't affect user ingestion at all
    adapter = get_adapter(redact_names_and_urls=True, strip_text_content=True)
    raw_branches = get_raw_branches()

    with patch(
        'jf_ingest.jf_git.clients.azure_devops.AzureDevopsClient.get_branches',
        return_value=raw_branches,
    ):
        standard_branches = adapter.get_branches_for_repo(BASIC_REPO, pull_branches=True)
        branch_names = []
        for standard_branch, raw_branch in zip(standard_branches, raw_branches):
            is_default = standard_branch.name == BASIC_REPO.default_branch_name
            assert standard_branch.is_default == is_default
            if standard_branch.is_default:
                assert standard_branch.name == BASIC_REPO.default_branch_name
            else:
                assert standard_branch.name.startswith('redacted-000')
            assert standard_branch.repo_id == BASIC_REPO.id
            assert standard_branch.sha == raw_branch['objectId']
            branch_names.append(standard_branch.name)

        # Assert uniqueness
        branch_names_are_unique = len(branch_names) == len(set(branch_names))
        assert branch_names_are_unique, f'List of Branch names is not unique! {branch_names}'


def test_get_commits_for_branches():
    # NOTE This test doesn't test if we're pulling for branches at all, it's just testing
    # that we are stripping text content!
    adapter = get_adapter(redact_names_and_urls=True, strip_text_content=True)
    raw_commits = get_raw_commits()
    with patch(
        'jf_ingest.jf_git.clients.azure_devops.AzureDevopsClient.get_commits',
        return_value=raw_commits,
    ):
        standard_commits = adapter.get_commits_for_branches(
            standardized_repo=BASIC_REPO, branches=[]
        )
        for raw_commit, standard_commit in zip(raw_commits, standard_commits):
            assert standard_commit.hash == raw_commit['commitId']
            assert standard_commit.url == None
            assert standard_commit.message == GitAdapter.sanitize_text(raw_commit['comment'], True)
            assert standard_commit.branch_name == BASIC_REPO.default_branch_name
            assert standard_commit.commit_date == parser.parse(raw_commit['committer']['date'])
            assert standard_commit.author_date == parser.parse(raw_commit['author']['date'])
            assert type(standard_commit.author) == StandardizedUser
            assert standard_commit.author != None
            assert standard_commit.repo.id == BASIC_REPO.id
            assert standard_commit.repo.url == BASIC_REPO.url
            assert standard_commit.repo.name == BASIC_REPO.name
            assert standard_commit.is_merge == False


def test_get_pr_updated_date_only_closed():
    adapter = get_adapter()

    closed_api_prs = [datetime(2020, 1, 1), datetime(2024, 1, 1, tzinfo=timezone.utc)]
    closed_api_prs_raw = [
        {'closedDate': closed_api_pr.isoformat()} for closed_api_pr in closed_api_prs
    ]

    for closed_api_pr, closed_api_pr_raw in zip(closed_api_prs, closed_api_prs_raw):
        assert adapter.get_pr_updated_date(BASIC_REPO, closed_api_pr_raw) == closed_api_pr


def test_get_pr_updated_date_with_iterations():
    adapter = get_adapter()
    raw_iterations = get_raw_iterations()
    raw_prs = [
        {
            'pullRequestId': 0,
            'creationDate': datetime(2000, 10, 26, tzinfo=timezone.utc).isoformat(),
        },
        {
            'pullRequestId': 1,
            'creationDate': datetime(2000, 11, 26, tzinfo=timezone.utc).isoformat(),
            'closedDate': datetime(2000, 11, 26, tzinfo=timezone.utc).isoformat(),
        },
        {
            'pullRequestId': 2,
            'creationDate': datetime(2000, 11, 30, tzinfo=timezone.utc).isoformat(),
            'closedDate': datetime(2000, 11, 30, tzinfo=timezone.utc).isoformat(),
        },
    ]
    with patch(
        'jf_ingest.jf_git.clients.azure_devops.AzureDevopsClient.get_pull_request_iterations',
        return_value=raw_iterations,
    ):
        for pr in raw_prs:
            if closed_date := pr.get('closedDate'):
                closed_date = datetime.fromisoformat(closed_date)
                assert adapter.get_pr_updated_date(BASIC_REPO, pr) == closed_date
            else:
                latest_iteration_date_str = raw_iterations[-1]['updatedDate']
                latest_iteration_update = datetime.fromisoformat(latest_iteration_date_str)
                assert adapter.get_pr_updated_date(BASIC_REPO, pr) == latest_iteration_update


def test_git_provider_pr_endpoint_supports_date_filtering():
    adapter = get_adapter()
    assert adapter.git_provider_pr_endpoint_supports_date_filtering()


@contextmanager
def _get_pr_mocking_context_manager():
    raw_prs = get_raw_prs()
    raw_iterations = get_raw_iterations()
    raw_diffs = get_raw_diffs()
    raw_change_counts = get_raw_change_counts()
    raw_commits = get_raw_commits()
    raw_threads = get_raw_threads_comments()
    with patch(
        'jf_ingest.jf_git.clients.azure_devops.AzureDevopsClient.get_pull_requests',
        return_value=raw_prs,
    ), patch(
        'jf_ingest.jf_git.clients.azure_devops.AzureDevopsClient.get_pull_request_iterations',
        return_value=raw_iterations,
    ), patch(
        'jf_ingest.jf_git.clients.azure_devops.AzureDevopsClient.get_pull_request_diff',
        return_value=raw_diffs,
    ), patch(
        'jf_ingest.jf_git.clients.azure_devops.AzureDevopsClient.get_pull_request_changes_counts',
        return_value=raw_change_counts,
    ), patch(
        'jf_ingest.jf_git.clients.azure_devops.AzureDevopsClient.get_pr_commits',
        return_value=raw_commits,
    ), patch(
        'jf_ingest.jf_git.clients.azure_devops.AzureDevopsClient.get_pr_comment_threads',
        return_value=raw_threads,
    ):
        yield raw_prs, raw_iterations, raw_diffs, raw_change_counts, raw_commits, raw_threads


def _test_get_prs_helper(
    redact_names_and_urls: bool,
    strip_text_content: bool,
    pull_files_for_prs: bool,
    hash_files_for_prs: bool,
):
    adapter = get_adapter(
        redact_names_and_urls=redact_names_and_urls, strip_text_content=strip_text_content
    )

    with _get_pr_mocking_context_manager() as raw_values:
        raw_prs, raw_iterations, raw_diffs, raw_change_counts, raw_commits, raw_threads = raw_values
        standard_prs = adapter.get_prs(
            standardized_repo=BASIC_REPO,
            pull_files_for_pr=pull_files_for_prs,
            hash_files_for_prs=hash_files_for_prs,
        )
        for raw_pr, standard_pr in zip(raw_prs, standard_prs):
            assert standard_pr.id == raw_pr['pullRequestId']
            assert standard_pr.additions == raw_change_counts['Add']
            assert standard_pr.deletions == raw_change_counts['Delete']
            assert standard_pr.changed_files == sum(raw_change_counts.values())
            assert standard_pr.is_closed == (raw_pr['status'] != 'active')
            assert standard_pr.is_merged == (raw_pr['status'] == 'completed')
            if standard_pr.is_closed:
                assert standard_pr.closed_date == parser.parse(raw_pr['closedDate'])
                assert standard_pr.updated_at == parser.parse(raw_pr['closedDate'])
            else:
                assert standard_pr.closed_date == None
                assert standard_pr.updated_at == parser.parse(raw_iterations[-1]['updatedDate'])
            if standard_pr.is_merged:
                assert standard_pr.merged_by != None
                assert type(standard_pr.merged_by) is StandardizedUser
                assert standard_pr.merge_date == parser.parse(raw_pr['mergeDate'])
                assert standard_pr.is_closed == True
                assert standard_pr.updated_at == standard_pr.closed_date
                assert standard_pr.merge_commit != None
                assert type(standard_pr.merge_commit) is StandardizedCommit
            else:
                assert standard_pr.merge_commit == None
                assert standard_pr.merged_by == None
                assert standard_pr.merge_date == None
                assert standard_pr.updated_at == parser.parse(raw_iterations[-1]['updatedDate'])
                assert standard_pr.is_closed == False
            assert standard_pr.created_at == parser.parse(raw_pr['creationDate'])
            assert standard_pr.title == GitAdapter.sanitize_text(
                raw_pr['title'], strip_text_content
            )
            assert standard_pr.body == GitAdapter.sanitize_text(
                raw_pr['description'], strip_text_content
            )
            assert standard_pr.url == (raw_pr['url'] if not redact_names_and_urls else None)
            if not redact_names_and_urls:
                assert standard_pr.head_branch == raw_pr['sourceRefName'].replace('refs/heads/', '')
                assert standard_pr.base_branch == raw_pr['targetRefName'].replace('refs/heads/', '')
            else:
                if standard_pr.head_branch != 'master':
                    assert 'redacted' in standard_pr.head_branch
                if standard_pr.base_branch != 'master':
                    assert 'redacted' in standard_pr.base_branch
            assert standard_pr.base_repo == BASIC_REPO
            assert standard_pr.head_repo == BASIC_REPO

            assert standard_pr.author != None
            assert type(standard_pr.author) is StandardizedUser

            files = [raw_file['item']['path'] for raw_file in raw_diffs]
            if hash_files_for_prs:
                files = [hash_filename(file) for file in files]
            assert standard_pr.files == (files if pull_files_for_prs else [])
            labels = [label['name'] for label in raw_pr.get('labels', [])]
            assert standard_pr.labels == labels

            # Assert we're getting all comments
            comments = []
            for thread in raw_threads:
                comments.extend(thread['comments'])

            assert len(standard_pr.comments) == len(comments)
            assert len(standard_pr.approvals) == 4

            assert len(standard_pr.commits) == len(raw_commits)
            if redact_names_and_urls:
                for commit in standard_pr.commits:
                    assert commit.url == None
                    assert 'redacted' in commit.branch_name


def test_get_prs_base():
    _test_get_prs_helper(
        redact_names_and_urls=False,
        strip_text_content=False,
        pull_files_for_prs=False,
        hash_files_for_prs=False,
    )


def test_get_prs_redact_names_and_urls():
    _test_get_prs_helper(
        redact_names_and_urls=True,
        strip_text_content=False,
        pull_files_for_prs=False,
        hash_files_for_prs=False,
    )


def test_get_prs_strip_text_content():
    _test_get_prs_helper(
        redact_names_and_urls=False,
        strip_text_content=True,
        pull_files_for_prs=False,
        hash_files_for_prs=False,
    )


def test_get_prs_pull_files():
    _test_get_prs_helper(
        redact_names_and_urls=False,
        strip_text_content=False,
        pull_files_for_prs=True,
        hash_files_for_prs=False,
    )


def test_get_prs_hash_files():
    _test_get_prs_helper(
        redact_names_and_urls=False,
        strip_text_content=False,
        pull_files_for_prs=True,
        hash_files_for_prs=True,
    )


def test_get_prs_all_true_variables():
    _test_get_prs_helper(
        redact_names_and_urls=True,
        strip_text_content=True,
        pull_files_for_prs=True,
        hash_files_for_prs=True,
    )



import json
import logging
import time
import traceback
import zoneinfo
from dataclasses import dataclass
from datetime import datetime
from types import MappingProxyType

from jira import JIRAError

from jf_ingest.config import JiraDownloadConfig
from jf_ingest.data_manifests.jira.adapters.manifest_adapter import ManifestAdapter
from jf_ingest.data_manifests.manifest_base import ManifestSource
from jf_ingest.jf_jira import get_jira_connection
from jf_ingest.jf_jira.downloaders import search_users
from jf_ingest.utils import get_wait_time, retry_for_status

logger = logging.getLogger(__name__)


@dataclass
class JiraCloudManifestAdapter(ManifestAdapter):
    '''
    Adapter for retrieving information from all remote Jira sources
    '''

    def __init__(
        self,
        config: JiraDownloadConfig,
        project_keys_to_classification_type: dict[str, str] = MappingProxyType({}),
    ):
        # Super class fields
        super().__init__(
            config=config,
            manifest_source=ManifestSource.remote,
            project_keys_to_classification_type=project_keys_to_classification_type,
        )
        self.jira_connection = get_jira_connection(config=config)

        # Rip off trailing slash since all urls depend on that
        self.jira_endpoint = (
            self.config.url[:-1] if self.config.url.endswith('/') else self.config.url
        )

        self._boards_cache = []
        self._projects_cache = []

        # Set up caches
        self._get_all_boards()
        self._get_all_projects()

    def get_users_count(self) -> int:
        # Users are a special case where their page results aren't formatted the same as
        # other API page results. Special, internal function to handle this
        def _get_user_count_from_pages(url_path: str, page_size: int = 50) -> int:
            curr = 0
            user_count = 0
            while True:
                page = retry_for_status(
                    self.jira_connection._get_json,
                    url_path,
                    {**{'startAt': curr, "maxResults": page_size}},
                )
                page_count = len(page)
                user_count += page_count
                curr += page_size

                if page_count == 0:
                    break
            return user_count

        # Jira Cloud
        if self.config.gdpr_active:
            return _get_user_count_from_pages(url_path='users/search')
        # Jira Server
        else:
            return len(
                search_users(
                    jira_connection=self.jira_connection,
                    gdpr_active=self.config.gdpr_active,
                    search_users_by_letter_email_domain=self.config.search_users_by_letter_email_domain,
                    required_email_domains=self.config.required_email_domains,
                    is_email_required=self.config.is_email_required,
                )
            )

    def get_fields_count(self) -> int:
        # Lazy loading paranoia, we might not need to do this for loop
        return len(retry_for_status(self.jira_connection.fields))

    def get_resolutions_count(self) -> int:
        # Lazy loading paranoia, we might not need to do this for loops
        try:
            return len(retry_for_status(self.jira_connection.resolutions))
        except Exception as e:
            logger.error(f"Could not get resolutions count for {self.config}, got {e}")
            return 0

    def get_issue_types_count(self) -> int:
        # Lazy loading paranoia, we might not need to do this for loop
        return len(retry_for_status(self.jira_connection.issue_types))

    def get_issue_link_types_count(self) -> int:
        return len(retry_for_status(self.jira_connection.issue_link_types))

    def get_priorities_count(self) -> int:
        return len(retry_for_status(self.jira_connection.priorities))

    def get_boards_count(self) -> int:
        return len(retry_for_status(self._get_all_boards))

    def get_project_versions_count(self) -> int:
        total_project_versions = 0
        retries = 0
        max_retries = 5
        constant_sleep = 5
        for project in self._get_all_projects():
            done = False
            while not done:
                try:
                    total_project_versions += self._get_raw_result(
                        url=f'{self.jira_endpoint}/rest/api/latest/project/{project["id"]}/version?startAt=0&maxResults=0&state=future,active,closed'
                    )['total']
                    retries = 0
                    done = True
                except Exception as e:
                    if (
                        hasattr(e, 'status_code')
                        and e.status_code == 429
                        and retries <= max_retries
                    ):
                        sleep_secs = get_wait_time(e, retries)
                        logger.info(
                            f'Caught JIRAError with a {e.status_code} return code during get_project_versions_count. '
                            f'Hopefully transient; sleeping for {sleep_secs} secs then may retry ({retries} of {max_retries}).'
                        )
                        time.sleep(sleep_secs + (constant_sleep * retries))
                        retries += 1
                    else:
                        logger.error(
                            f"Could not recover from error getting jira projects for manifests, got {e}"
                        )
                        raise

        return total_project_versions

    def _get_all_projects(self) -> list[dict]:
        if not self._projects_cache:
            if self.config.gdpr_active:
                self._projects_cache = [
                    project
                    for project in self._page_get_results(
                        url=f'{self.jira_endpoint}/rest/api/latest/project/search?startAt=%s&maxResults=500&status=live'
                    )
                    if not project.get('archived', False) and not project['isPrivate']
                ]
            else:
                self._projects_cache = [
                    project
                    for project in self._get_raw_result(
                        url=f'{self.jira_endpoint}/rest/api/latest/project?includedArchived=True'
                    )
                ]

        return self._projects_cache

    def get_project_data_dicts(self) -> list[dict]:
        return [
            {"key": project['key'], "id": project['id']} for project in self._get_all_projects()
        ]

    def get_project_versions_count_for_project(self, project_id: int) -> int:
        return self._get_raw_result(
            url=f'{self.jira_endpoint}/rest/api/latest/project/{project_id}/version?startAt=0&maxResults=0&state=future,active,closed'
        )['total']

    def get_issues_count(self) -> int:
        # JCR 072123 Since we're using JiraIssueData, we need to check `download_all_json`, since this
        # determines whether we get _all_ issue data or just data from the pull_from date.
        #
        # See: `download_all_issue_metadata`
        pull_from = (
            '0' if self.config.full_redownload else self.config.pull_from.strftime('%Y-%m-%d')
        )

        # Query for all issues via JQL, but ask for 0 results.
        # Meta data is returned that will give us the total number of
        # (unpulled) results
        result = self._get_jql_search(jql_search=f"updatedDate > {pull_from}", max_results=0)
        return result['total']

    def get_board_ids(self, project_id: int):
        result = self._get_raw_result(
            url=f'{self.jira_endpoint}/rest/agile/1.0/board?projectKeyOrId={project_id}&startAt=0&maxResults=100'
        )
        return [value['id'] for value in result['values']]

    def _get_all_boards(self):
        if not self._boards_cache:
            self._boards_cache = [
                board
                for board in self._page_get_results(
                    url=f'{self.jira_endpoint}/rest/agile/1.0/board?startAt=%s&maxResults=100'
                )
            ]

        return self._boards_cache

    def test_basic_auth_for_project(self, project_id: int) -> bool:
        # Doing a basic query for issues is the best way to test auth.
        # Catch and error, if it happens, and bubble up more specific error
        try:
            self.get_issues_count_for_project(project_id=project_id)
            return True
        except JIRAError:
            return False
        except Exception:
            # This is unexpected behavior and it should never happen, log the error
            # before returning
            logger.error(
                'Unusual exception encountered when testing auth. '
                f'JIRAError was expected but the following error was raised: {traceback.format_exc()}'
            )
            return False

    def get_issues_count_for_project(self, project_id: int) -> int:
        # JCR 072123 Since we're using JiraIssueData, we need to check `download_all_json`, since this
        # determines whether we get _all_ issue data or just data from the pull_from date.
        #
        # See: `download_all_issue_metadata`
        pull_from = (
            '0' if self.config.full_redownload else self.config.pull_from.strftime('%Y-%m-%d')
        )

        # Query for all issues via JQL, but ask for 0 results.
        # Meta data is returned that will give us the total number of
        # (unpulled) results
        return self._get_jql_search(
            jql_search=f"project = {project_id} AND updatedDate > {pull_from}",
            max_results=0,
        )['total']

    # I believe that there should be a one to one relationship with
    # Jira Issues and Jira Issue Data...
    def get_issues_data_count_for_project(self, project_id: int) -> int:
        return self.get_issues_count_for_project(project_id=project_id)

    def _get_jql_search(self, jql_search: str, max_results: int = 0):
        retries = 0
        max_retries = 5
        constant_sleep = 5
        while True:
            try:
                return retry_for_status(
                    self.jira_connection._get_json,
                    'search',
                    {'jql': jql_search, "maxResults": max_results},
                )
            except Exception as e:
                if hasattr(e, 'status_code') and e.status_code == 429 and retries <= max_retries:
                    sleep_secs = get_wait_time(e, retries)
                    logger.info(
                        f'Caught JIRAError with a {e.status_code} return code during jira_connection._get_json. '
                        f'Hopefully transient; sleeping for {sleep_secs} secs then may retry ({retries} of {max_retries}).'
                    )
                    time.sleep(sleep_secs + (constant_sleep * retries))
                    retries += 1
                else:
                    logger.error(
                        f"Could not recover from error getting jira projects for manifests, got {e}"
                    )
                    raise

    def _page_get_results(self, url: str):
        start_at = 0
        while True:
            page_result = self._get_raw_result(url % start_at)
            for value in page_result['values']:
                yield value

            if page_result['isLast']:
                break
            else:
                start_at += len(page_result['values'])

    def _get_raw_result(self, url) -> dict:
        response = retry_for_status(self.jira_connection._session.get, url)
        response.raise_for_status()
        json_str = response.content.decode()
        return json.loads(json_str)

    def get_last_updated_for_project(self, project_id: int) -> datetime:
        try:
            most_recent_issue_from_api: dict = self._get_jql_search(
                jql_search=f"project = {project_id} ORDER BY updated",
                max_results=1,
            )['issues'][0]
            # Example returned str: '2023-05-15T14:04:19.376-0400'
            most_recent_updated_date_str = most_recent_issue_from_api['fields']['updated']
            last_updated_datetime = datetime.strptime(
                most_recent_updated_date_str, '%Y-%m-%dT%H:%M:%S.%f%z'
            )
            # Normalize to utc
            return last_updated_datetime.astimezone(tz=zoneinfo.ZoneInfo('UTC'))
        except Exception as e:
            logger.error(
                f"Could not fetch latest issue updated date in project {project_id}, got: {e}"
            )
            return None

    """
    NOTE: DEPRECATED FUNCTION
    This function takes a long time to process and we currently do not use sprint counts for anything.
    I have made the decision to deprecate it, but there is no standard decorator for deprecated functions in python.
    Instead, I have removed this function from the parent abstract class and commented it out.
    I have opted to comment it out because it's a non trivial function, and if we want sprint counts back
    sometime in the future, it'd be nice to have this to work off of
    -Gavin
    
    NOTE: If we do bring this back, we must add retry logic with exponential back off logic to get around 429 errors!
    
    def get_sprints_count(self) -> int:
        # Leverage the jira_connections _session field to get an authenticated session
        # Get all boards with above query, get all sprints with this query: https://docs.atlassian.com/jira-software/REST/7.3.1/?#agile/1.0/board/{boardId}/sprin

        def _get_sprints_for_board(board_id: int):
            total_sprints_for_board = 0
            try:
                total_sprints_for_board += len(
                    [
                        sprint
                        for sprint in self._page_get_results(
                            url=f'{self.jira_endpoint}/rest/agile/1.0/board/{board_id}/sprint?startAt=%s&maxResults=100'
                        )
                    ]
                )
            except JIRAError as e:
                # JIRA returns 500 errors for various reasons: board is
                # misconfigured; "failed to execute search"; etc.  Just
                # skip and move on
                if e.status_code == 500:
                    logger.debug(f"Couldn't get sprints for {self.instance} board {board_id}")
                elif e.status_code == 400:
                    logger.debug(
                        f"For {self.instance}, board {board_id} doesn't support sprints -- skipping"
                    )
                else:
                    raise

            return total_sprints_for_board

        with ThreadPoolExecutor(max_workers=5) as executor:
            futures = [
                executor.submit(_get_sprints_for_board, board["id"])
                for board in self._get_all_boards()
            ]

            return sum([future.result() for future in futures if future])
    """



import json
import os
import queue
from copy import deepcopy
from unittest.mock import patch

import pytest
import requests
import requests_mock

from jf_ingest.config import IngestionConfig, IngestionType, JiraDownloadConfig
from jf_ingest.jf_jira.custom_fields import (
    JELLYFISH_CUSTOM_FIELDS_API_LIMIT,
    _annotate_results_from_jellyfish,
    _retrieve_custom_fields_from_jellyfish,
)
from jf_ingest.utils import RetryLimitExceeded
from tests.jf_jira.utils import get_fixture_file_data

ATLASSIAN_BASE_URL = 'mock://test-co.atlassian.net/'
JELLYFISH_BASE_URL = 'mock://jellyfish.jellyfish'
JELLYFISH_CUSTOM_FIELDS_ENDPOINT = 'endpoints/jira/issues/custom-fields'


def _get_ingestion_config():
    company_slug = "test_company"
    url = ATLASSIAN_BASE_URL
    return IngestionConfig(
        company_slug=company_slug,
        jira_config=JiraDownloadConfig(
            company_slug=company_slug,
            url=url,
            personal_access_token='pat',
        ),
        git_configs=[],
        jellyfish_api_token='some_token',
        jellyfish_api_base=JELLYFISH_BASE_URL,
        ingest_type=IngestionType.DIRECT_CONNECT,
    )


def _get_jellyfish_custom_fields_default_response() -> dict:
    return json.loads(get_fixture_file_data(fixture_path='jellyfish_custom_fields.json'))


def _get_jellyfish_custom_fields_empty_response() -> dict:
    return json.loads(
        get_fixture_file_data(fixture_path='jellyfish_custom_fields_empty_response.json')
    )


def test_custom_fields_base_case():
    config = _get_ingestion_config()
    jf_response = _get_jellyfish_custom_fields_default_response()
    empty_response = _get_jellyfish_custom_fields_empty_response()

    with requests_mock.Mocker() as mocker:
        base_mock = mocker.register_uri(
            'GET',
            f"{config.jellyfish_api_base}/{JELLYFISH_CUSTOM_FIELDS_ENDPOINT}?offset=0&limit={JELLYFISH_CUSTOM_FIELDS_API_LIMIT}",
            json=jf_response,
            status_code=200,
        )

        addl_mock = mocker.register_uri(
            'GET',
            f"{config.jellyfish_api_base}/{JELLYFISH_CUSTOM_FIELDS_ENDPOINT}?offset={len(jf_response['issues'])}&limit={JELLYFISH_CUSTOM_FIELDS_API_LIMIT}",
            json=empty_response,
            status_code=200,
        )

        q = queue.Queue()
        _retrieve_custom_fields_from_jellyfish(config, q)
        assert q.qsize() == 1
        output = q.get()

        assert output == _annotate_results_from_jellyfish(jf_response)
        assert addl_mock.call_count == 0
        assert base_mock.call_count == 1
        assert base_mock.last_request.qs == {
            'offset': ['0'],
            'limit': [str(JELLYFISH_CUSTOM_FIELDS_API_LIMIT)],
        }


def test_custom_fields_api_adjusted_limit():
    config = _get_ingestion_config()
    jf_response = _get_jellyfish_custom_fields_default_response()
    empty_response = _get_jellyfish_custom_fields_empty_response()

    jf_response['max_records'] = 1
    jf_response['total_records'] = 1
    jf_response_1 = deepcopy(jf_response)
    jf_response_2 = deepcopy(jf_response)
    jf_response_1['issues'] = jf_response_1['issues'][:1]
    jf_response_2['issues'] = jf_response_2['issues'][1:]

    with requests_mock.Mocker() as mocker:
        general_mock = mocker.register_uri(
            'GET',
            f"{config.jellyfish_api_base}/{JELLYFISH_CUSTOM_FIELDS_ENDPOINT}",
            json=empty_response,
            status_code=200,
        )

        mock0 = mocker.register_uri(
            'GET',
            f"{config.jellyfish_api_base}/{JELLYFISH_CUSTOM_FIELDS_ENDPOINT}?offset=0&limit={JELLYFISH_CUSTOM_FIELDS_API_LIMIT}",
            json=jf_response_1,
            status_code=200,
        )

        mock1 = mocker.register_uri(
            'GET',
            f"{config.jellyfish_api_base}/{JELLYFISH_CUSTOM_FIELDS_ENDPOINT}?offset=1&limit=1",
            json=jf_response_2,
            status_code=200,
        )

        q = queue.Queue()
        _retrieve_custom_fields_from_jellyfish(config, q)
        assert q.qsize() == 2
        output0 = q.get()
        output1 = q.get()

        assert output0 == _annotate_results_from_jellyfish(jf_response_1)
        assert output1 == _annotate_results_from_jellyfish(jf_response_2)
        assert mock0.call_count == 1
        assert mock1.call_count == 1
        assert general_mock.call_count == 1


def test_custom_fields_caller_adjusted_limit():
    config = _get_ingestion_config()
    jf_response = _get_jellyfish_custom_fields_default_response()
    empty_response = _get_jellyfish_custom_fields_empty_response()
    limit = 3

    with requests_mock.Mocker() as mocker:
        general_mock = mocker.register_uri(
            'GET',
            f"{config.jellyfish_api_base}/{JELLYFISH_CUSTOM_FIELDS_ENDPOINT}",
            json=empty_response,
            status_code=200,
        )

        mock0 = mocker.register_uri(
            'GET',
            f"{config.jellyfish_api_base}/{JELLYFISH_CUSTOM_FIELDS_ENDPOINT}?offset=0&limit={limit}",
            json=jf_response,
            status_code=200,
        )

        q = queue.Queue()
        _retrieve_custom_fields_from_jellyfish(config, q, limit=limit)
        assert q.qsize() == 1
        output0 = q.get()

        assert output0 == _annotate_results_from_jellyfish(jf_response)
        assert mock0.call_count == 1
        assert general_mock.call_count == 0


def test_custom_fields_max_issues_to_process():
    config = _get_ingestion_config()
    jf_response = _get_jellyfish_custom_fields_default_response()
    empty_response = _get_jellyfish_custom_fields_empty_response()

    jf_response['max_records'] = 1
    jf_response['total_records'] = 1
    jf_response['issues'] = jf_response['issues'][:1]

    with requests_mock.Mocker() as mocker:
        general_mock = mocker.register_uri(
            'GET',
            f"{config.jellyfish_api_base}/{JELLYFISH_CUSTOM_FIELDS_ENDPOINT}",
            json=empty_response,
            status_code=200,
        )

        mock0 = mocker.register_uri(
            'GET',
            f"{config.jellyfish_api_base}/{JELLYFISH_CUSTOM_FIELDS_ENDPOINT}?offset=0&limit=1",
            json=jf_response,
            status_code=200,
        )

        q = queue.Queue()
        _retrieve_custom_fields_from_jellyfish(config, q, max_issues_to_process=1)
        assert q.qsize() == 1
        output0 = q.get()

        assert output0 == _annotate_results_from_jellyfish(jf_response)
        assert mock0.call_count == 1
        assert general_mock.call_count == 0


def test_custom_fields_non_handled_status_code():
    config = _get_ingestion_config()

    with requests_mock.Mocker() as mocker:
        general_mock = mocker.register_uri(
            'GET',
            f"{config.jellyfish_api_base}/{JELLYFISH_CUSTOM_FIELDS_ENDPOINT}",
            json={},
            status_code=418,
        )

        q = queue.Queue()
        with pytest.raises(requests.exceptions.HTTPError):
            _retrieve_custom_fields_from_jellyfish(config, q, max_issues_to_process=1)


def test_custom_fields_retry_for_status():
    config = _get_ingestion_config()

    with requests_mock.Mocker() as mocker:
        general_mock = mocker.register_uri(
            'GET',
            f"{config.jellyfish_api_base}/{JELLYFISH_CUSTOM_FIELDS_ENDPOINT}",
            json={},
            status_code=504,
        )

        q = queue.Queue()
        with patch("jf_ingest.utils.time.sleep", return_value=0) as m:
            with pytest.raises(RetryLimitExceeded):
                _retrieve_custom_fields_from_jellyfish(config, q, max_issues_to_process=1)
